@@ -119,19 +119,22 @@ class CocciMatch:
         self.column = int(mcol)
         self.lineend = int(mlineend)
         self.columnend = int(mcolend)
+        self.start_at = self.line
+        self.stop_at = self.line
+        self.trailer = ""
 
-    def display(self, stype, mode='raw', oformat='term', before=0, after=0):
+    def display(self, stype, mode='raw', oformat='term'):
         """
         Display output for a single match
 
+        :param stype: name of the matched type
+        :type stype: str
         :param mode: display mode
         :type mode: str
         :param oformat: format of output for color (term, html)
         :type oformat: str
-        :param before: number of lines to display before match
-        :type before: int
-        :param after: number of lines to display after match
-        :type after: int
+        :return: a human readable string containing the result of the search
+                 (matched line, context, file name, etc.)
         """
         f = open(self.file, 'r')
         lines = f.readlines()
@@ -142,19 +145,23 @@ class CocciMatch:
         output = ""
         if mode == 'color':
             output += "%s: l.%s -%d, l.%s +%d, %s %s%s\n" % (self.file,
-                 self.line, before, self.line, after, stype, ptype, pmatch)
-        for i in range(int(self.line) - 1 - before, int(self.line) + after):
+                 self.line, self.line - self.start_at, self.line,
+                 self.stop_at - self.line, stype, ptype, pmatch)
+        for i in range(self.start_at - 1, min(self.stop_at, len(lines))):
             if mode == 'color':
                 output += lines[i]
             elif mode == 'vim':
-                output += "%s|%s| (%s %s%s): %s" % (self.file, self.line,
+                output += "%s|%s| (%s %s%s): %s" % (self.file, i + 1,
                 stype, ptype, pmatch, lines[i])
             elif mode == 'emacs':
-                output += "%s:%s: (%s %s%s): %s" % (self.file, self.line,
+                output += "%s:%s: (%s %s%s): %s" % (self.file, i + 1,
+                stype, ptype, pmatch, lines[i])
+            elif i == self.line - 1:
+                output += "%s:%s (%s %s%s): %s" % (self.file, i + 1,
                 stype, ptype, pmatch, lines[i])
             else:
-                output += "%s:%s (%s %s%s): %s" % (self.file, self.line,
-                stype, ptype, pmatch, lines[i])
+                output += "%s-%s %s - %s" % (self.file, i + 1,
+                ' ' * (2 + len(stype + ptype + pmatch)), lines[i])
         f.close()
         if mode == 'color':
             if have_pygments:
@@ -168,7 +175,7 @@ class CocciMatch:
                         HtmlFormatter(noclasses=True))
                 else:
                     return output
-        return output
+        return output + self.trailer
 
 
 class CocciProcess:
@@ -300,13 +307,13 @@ for p in p1:
             op = _operation_name(fname)
             self.operations[op] = path.join(self.get_datadir(), fname)
 
-    def setup(self, stype, attribut, operation):
+    def setup(self, stype, attribute, operation):
         """
         :param stype: structure name, used to replace '$type' in the cocci file
         :type stype: str
-        :param attribut: basically attribut of the structure, used to replace
-            '$attribut' in the cocci file
-        :type attribut: str
+        :param attribute: basically attribute of the structure, used to replace
+            '$attribute' in the cocci file
+        :type attribute: str
         :param operation: search operation to do
         :type operation: str
         :raise: :class:`CocciRunException`
@@ -315,7 +322,7 @@ for p in p1:
             raise CocciRunException("Can't use coccigrep without type to "
                 "search")
         self.type = stype
-        self.attribut = attribut
+        self.attribute = attribute
         self.operation = operation
 
     def set_concurrency(self, ncpus):
@@ -397,27 +404,24 @@ for p in p1:
 
     def run(self, files):
         """
-        Run the search against the files given in argument
+        Run the search against the files and directories given in argument
 
         This function is doing the main job. It will run spatch with
         the correct parameters by using subprocess or it will use
         multiprocessing if a concurrency level greater than 1 has been
         asked.
 
-        :param files: list of filenames
-        :type files: list of str
+        :param args: list of filenames and directory names
+        :type args: list of str
         :raise: :class:`CocciRunException` or :class:`CocciConfigException`
         """
 
         if len(files) == 0:
             raise CocciRunException("Can't use coccigrep without files "
                 "to search")
-        for cfile in files:
-            if not path.isfile(cfile):
-                raise CocciRunException("'%s' is not a file, can't "
-                    "continue" % cfile)
+
         # create tmp cocci file:
-        tmp_cocci_file = NamedTemporaryFile(suffix=".cocci", delete=False)
+        tmp_cocci_file = NamedTemporaryFile(suffix=".cocci")
         tmp_cocci_file_name = tmp_cocci_file.name
         # open file with name matching operation
         cocci_file = open(self.operations[self.operation], 'r')
@@ -426,11 +430,11 @@ for p in p1:
         cocci_smpl_tmpl = Template(cocci_tmpl)
         cocci_file.close()
         cocci_smpl = cocci_smpl_tmpl.substitute(type=self.type,
-            attribut=self.attribut)
+            attribute=self.attribute)
         cocci_grep = cocci_smpl + CocciGrep.cocci_python
 
         tmp_cocci_file.write(cocci_grep)
-        tmp_cocci_file.close()
+        tmp_cocci_file.flush()
 
         # launch spatch
         output = ""
@@ -462,8 +466,8 @@ for p in p1:
                     continue
                 import pickle
                 err = pickle.loads(ret)
-                unlink(tmp_cocci_file_name)
                 _raise_run_err(err, cmd)
+            tmp_cocci_file.close()
         # Fallback to one spatch
         else:
             cmd = [self.spatch]
@@ -480,7 +484,7 @@ for p in p1:
                 unlink(tmp_cocci_file_name)
                 _raise_run_err(err, cmd)
 
-            unlink(tmp_cocci_file_name)
+            tmp_cocci_file.close()
 
         prevfile = None
         prevline = None
@@ -512,8 +516,30 @@ for p in p1:
         :type oformat: str
         :return: the result of the search as a str
         """
-        output = ''.join(match.display(self.type, mode=mode, oformat=oformat,
-            before=before, after=after)
+        if before != 0 or after != 0:
+            prev_match = None
+            for index in xrange(len(self.matches)):
+                cur_match = self.matches[index]
+                cur_match.start_at = cur_match.line - before
+                cur_match.stop_at = cur_match.line + after
+                if cur_match.start_at < 1:
+                    cur_match.start_at = 1
+
+                if prev_match is not None:
+                    prev_match.trailer = "--\n"
+                    if prev_match.file == cur_match.file:
+                        if prev_match.stop_at >= cur_match.line:
+                            prev_match.stop_at = cur_match.line - 1
+                        if prev_match.stop_at >= cur_match.start_at:
+                            cur_match.start_at = prev_match.stop_at + 1
+
+                        if prev_match.stop_at + 1 == cur_match.start_at:
+                            # No seprator if groups are contiguous
+                            prev_match.trailer = ""
+
+                prev_match = cur_match
+
+        output = ''.join(match.display(self.type, mode=mode, oformat=oformat)
             for match in self.matches)
 
         return output.rstrip()
